import numpy as np

## Electrochemistry constants
F = 96485 # C/mol, Faraday constant    
R = 8.315 # J/mol K, Gas constant
T = 298 # K, Temperature
FRT = F/(R*T)


class E:
    def __init__(self, wf, n=1, A=0.0314, E0=0, cOb=0, cRb=1e-6, 
                 DO=1e-5, DR=1e-5, k0=1e8, alpha=0.5):
        self.wf = wf
        self.n = n
        self.A = A
        self.E0 = E0
        self.cOb = cOb
        self.cRb = cRb
        self.DO = DO
        self.DR = DR
        self.k0 = k0
        self.alpha = alpha
        self.DOR = DO/DR

    def grid(self):
        self.nT = np.size(self.wf.t)
        self.dT = 1/self.nT
        self.lamb = 0.45
        self.Xmax = 6*np.sqrt(self.nT*self.lamb)
        self.dX = np.sqrt(self.dT/self.lamb)
        self.nX = int(self.Xmax/self.dX)

        ## Discretisation of variables and initialisation
        if self.cRb == 0: # In case only O present in solution
            self.CR = np.zeros([self.nT,self.nX])
            self.CO = np.ones([self.nT,self.nX])
        else:
            self.CR = np.ones([self.nT,self.nX])
            self.CO = np.ones([self.nT,self.nX])*self.cOb/self.cRb

        self.X = np.linspace(0,self.Xmax,self.nX) # Discretisation of distance
        self.eps = (self.wf.E-self.E0)*self.n*FRT # adimensional potential waveform
        self.delta = np.sqrt(self.DR*self.wf.t[-1]) # cm, diffusion layer thickness
        self.K0 = self.k0*self.delta/self.DR # Normalised standard rate constant


    def run(self):
        self.grid()
        for k in range(1,self.nT):
            # Boundary condition, Butler-Volmer:
            CR1kb = self.CR[k-1,1]
            CO1kb = self.CO[k-1,1]
            self.CR[k,0] = (CR1kb + self.dX*self.K0*np.exp(
                            -self.alpha*self.eps[k])*(CO1kb + CR1kb/self.DOR))/(
                       1 + self.dX*self.K0*(np.exp((1-self.alpha)*self.eps[k])+\
                       np.exp(-self.alpha*self.eps[k])/self.DOR))
            self.CO[k,0] = CO1kb + (CR1kb - self.CR[k,0])/self.DOR

            self.CR[k,1:-1] = self.CR[k-1,1:-1] + self.lamb*(self.CR[k-1,2:]\
                            - 2*self.CR[k-1,1:-1] + self.CR[k-1,:-2])
            self.CO[k,1:-1] = self.CO[k-1,1:-1] + self.lamb*(self.CO[k-1, 2:]\
                            - 2*self.CO[k-1, 1:-1] + self.CO[k-1,:-2])

        # Denormalising:
        if self.cRb:
            I = -self.CR[:,2] + 4*self.CR[:,1] - 3*self.CR[:,0]
            D = self.DR
            c = self.cRb
        else: # In case only O present in solution
            I = self.CO[:,2] - 4*self.CO[:,1] + 3*self.CO[:,0]
            D = self.DO
            c = self.cOb
        self.i = self.n*F*self.A*D*c*I/(2*self.dX*self.delta)

        self.cR = self.CR*self.cRb
        self.cO = self.CO*self.cOb
        self.x = self.X*self.delta

            


if __name__ == '__main__':
    import softpotato as sp
    print('Running from main')

    # Potential waveform:
    wf = sp.technique.Sweep(Eini=-0.5, Efin=0.5, sr=0.1)

    # Electrochemical parameters:
    n = 1
    A = 0.0314
    E0 = 0
    cOb = 0
    cRb = 1e-6
    DO = 3.25e-5
    DR = 3.25e-5
    k0 = 1e8
    alpha = 0.5

    # Simulation:
    sim = E(wf, n, A, E0, cOb, cRb, DO, DR, k0, alpha)
    sim.run()
    
    # Plotting
    sp.plotting.plot(wf.t, wf.E, xlab='$t$ / s', ylab='$E$ / V', fig=1, show=0)
    sp.plotting.plot(wf.E, sim.i*1e6, ylab='$i$ / $\mu$A', fig=2, show=1)




'''
    Copyright (C) 2020 Oliver Rodriguez
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

    Created on Wed Jul 22 15:07:06 2020
    * R - e- -> O
    * Diffusion, E mechanism
    * Butler Volmer
    * Explicit finite differences

    @author: oliverrdz
    https://oliverrdz.xyz
'''

import numpy as np
import plots as p
import waveforms as wf

## Electrochemistry constants
F = 96485 # C/mol, Faraday constant    
R = 8.315 # J/mol K, Gas constant
T = 298 # K, Temperature
FRT = F/(R*T)

n=1
Ageo = 0.0314
cOb=0#1e-6
cRb=1e-6
DO=3.25e-5
DR=3.25e-5
ks=1e8
alpha=0.5

# Potential waveform
E0 = 0  # V, standard potential
Eini = -0.5 # V, initial potential
Efin = 0.5 # V, final potential vertex
sr = 0.1 # V/s, scan rate
ns = 2 # number of sweeps
dE = 0.01 # V, potential increment. This value has to be small for BI to approximate the circuit properly

t, E = wf.sweep(Eini=Eini, Efin=Efin, dE=dE, sr=sr, ns=ns)

DOR = DO/DR

#%% Simulation parameters
nT = np.size(t) # number of time elements
dT = 1/nT # adimensional step time
lamb = 0.45 # For the algorithm to be stable, lamb = dT/dX^2 < 0.5
Xmax = 6*np.sqrt(nT*lamb) # Infinite distance
dX = np.sqrt(dT/lamb) # distance increment
nX = int(Xmax/dX) # number of distance elements

## Discretisation of variables and initialisation
if cRb == 0: # In case only O present in solution
    CR = np.zeros([nT,nX])
    CO = np.ones([nT,nX])
else:
    CR = np.ones([nT,nX])
    CO = np.ones([nT,nX])*cOb/cRb


X = np.linspace(0,Xmax,nX) # Discretisation of distance
eps = (E-E0)*n*FRT # adimensional potential waveform
delta = np.sqrt(DR*t[-1]) # cm, diffusion layer thickness
K0 = ks*delta/DR # Normalised standard rate constant

#%% Simulation
for k in range(1,nT):
    # Boundary condition, Butler-Volmer:
    CR1kb = CR[k-1,1]
    CO1kb = CO[k-1,1]
    CR[k,0] = (CR1kb + dX*K0*np.exp(-alpha*eps[k])*(CO1kb + CR1kb/DOR))/(
               1 + dX*K0*(np.exp((1-alpha)*eps[k]) + np.exp(-alpha*eps[k])/DOR))
    CO[k,0] = CO1kb + (CR1kb - CR[k,0])/DOR

    # Solving finite differences:
    for j in range(1,nX-1):
        CR[k,j] = CR[k-1,j] + lamb*(CR[k-1,j+1] - 2*CR[k-1,j] + CR[k-1,j-1])
        CO[k,j] = CO[k-1,j] + DOR*lamb*(CO[k-1,j+1] - 2*CO[k-1,j] + CO[k-1,j-1])

# Denormalising:
if cRb:
    I = -CR[:,2] + 4*CR[:,1] - 3*CR[:,0]
    D = DR
    c = cRb
else: # In case only O present in solution
    I = CO[:,2] - 4*CO[:,1] + 3*CO[:,0]
    D = DO
    c = cOb
i = n*F*Ageo*D*c*I/(2*dX*delta)

cR = CR*cRb
cO = CO*cOb
x = X*delta

#%% Plot
p.plot(E, i*1e6, "$E$ / V", "$i$ / $\mu$A")
